# coding: utf-8
from django.contrib.sessions.backends.db import SessionBase

from EventAggregator.event_dispatcher import EventDispatcher


class EventAggregator:
    @staticmethod
    def publish(message, session: SessionBase = None):
        """
        :param message: 发布的消息
        :param session: 有值时，对应的响应(response)没有异常时发布; 为None立即发布
        :return:
        """
        if isinstance(session, SessionBase):
            if not hasattr(session, '_dispatch_events'):
                session._dispatch_events = []
            session._dispatch_events.append(lambda: EventDispatcher.dispatch_event(type(message), message))
        else:
            EventDispatcher.dispatch_event(type(message), message)

    @staticmethod
    def subscribe(message_class, handler, times=3):
        """
        :param message_class: 
        :param handler: 
        :param times: 失败时，重试的次数， 取值>=0 
        :return: 
        """
        handler._times = times
        EventDispatcher.add_event_listener(message_class, handler)

    @staticmethod
    def unsubscribe(message_class, handler):
        EventDispatcher.remove_event_listener(message_class, handler)



# coding: utf-8
import threading
import time

try:
    # needed to support Django >= 1.10 MIDDLEWARE
    from django.utils.deprecation import MiddlewareMixin
except ImportError:
    # needed to keep Django <= 1.9 MIDDLEWARE_CLASSES
    MiddlewareMixin = object


class EventPublishMiddleware(MiddlewareMixin):
    def __dispatch(self, dispatch_events):
        time.sleep(1)
        for dispatch_event in dispatch_events:
            if callable(dispatch_event):
                dispatch_event()

    def process_response(self, request, response):
        if int(response.status_code / 100) in [2, 3] and hasattr(request.session, '_dispatch_events'):
            dispatch_thread = threading.Thread(target=self.__dispatch, args=(request.session._dispatch_events,))
            dispatch_thread.setDaemon(True)
            dispatch_thread.start()
        return response



# coding: utf-8
import logging
import threading
import traceback

try:
    from django.db import close_old_connections
except ImportError:
    def close_old_connections():
        pass

LOGGER = logging.getLogger(__name__)


class EventDispatcher(object):
    def __init__(self):
        self._events = {}

    def __del__(self):
        self._events = None

    def __listener_executor(self, listener, args, kwargs):
        left_times = kwargs.pop('_left_times', 0)
        if left_times < 0:
            left_times = 0
        try:
            listener(*args, **kwargs)
        except:
            LOGGER.error({
                'listener': '%s.%s' % (getattr(listener, '__module__', 'None'), getattr(listener, '__name__', 'None')),
                'left_times(if > 0, will try in 100ms later)': left_times,
                'exceptions': traceback.format_exc(),
            })
            if left_times > 0:
                kwargs['_left_times'] = left_times - 1
                listener_thread = threading.Timer(0.1, function=self.__listener_executor, args=(listener, args, kwargs))
                listener_thread.setDaemon(True)
                listener_thread.start()
        finally:
            close_old_connections()

    def has_listener(self, event_type, listener):
        if event_type in self._events.keys():
            return listener in self._events[event_type]
        else:
            return False

    def dispatch_event(self, event_type, *args, **kwargs):
        if event_type in self._events.keys():
            listeners = self._events[event_type]

            for listener in listeners:
                times = getattr(listener, '_times', 3)
                kwargs['_left_times'] = kwargs.get('_left_times', times)
                listener_thread = threading.Thread(target=self.__listener_executor, args=(listener, args, kwargs))
                listener_thread.setDaemon(True)
                listener_thread.start()

    def add_event_listener(self, event_type, listener):
        if not self.has_listener(event_type, listener):
            listeners = self._events.get(event_type, [])
            listeners.append(listener)
            self._events[event_type] = listeners

    def remove_event_listener(self, event_type, listener):
        if self.has_listener(event_type, listener):
            listeners = self._events[event_type]
            if len(listeners) == 1:
                del self._events[event_type]
            else:
                listeners.remove(listener)
                self._events[event_type] = listeners


EventDispatcher = EventDispatcher()



import os
import sys
from distutils.core import setup


def read(fname):
    """
    define 'read' func to read long_description from 'README.txt'
    """
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name='Django-EventAggregator',
    version='1.4.0',
    author='songtao',
    author_email='975765671@qq.com',
    url='https://pypi.org/project/Django-EventAggregator/',
    license='MIT',
    description='Event Aggregator for Django project',
    long_description=read('README.txt'),
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
    ],
    keywords='django event',
    packages=['EventAggregator']
)

# command: python setup.py register sdist upload
